import torch
from torch import nn
import torch.nn.functional as F
from torch import optim
import torchvision
import torchvision.datasets as datasets
from torch.utils.data import TensorDataset
from torch.utils.data import DataLoader
import multiprocessing
from multiprocessing import Pool

import numpy
import math
import argparse
import os
import random
import model as model

parser = argparse.ArgumentParser(description='Toy')
parser.add_argument('--epochnumber', default=100000, type=int)
parser.add_argument('--initial', default=2, type=int)
parser.add_argument('--mininstance', default=1, type=int)
parser.add_argument('--instance', default=16, type=int)
parser.add_argument('--saveinst', default=0, type=int)
parser.add_argument('--batchsize', default=1000, type=int)
parser.add_argument('--testbatchsize', default=1000, type=int)
parser.add_argument('--minbatchsize', default=10, type=int)
parser.add_argument('--printstep', default=60, type=int)
parser.add_argument('--validbatch', default=100, type=int)
parser.add_argument('--lr', default=1e-2, type=float)
parser.add_argument('--loglr', default=11, type=float)
parser.add_argument('--annealratio', default=2**(1/10000), type=float)
parser.add_argument('--epochlen', default=1, type=int)
parser.add_argument('--xdim', default=784, type=int)
parser.add_argument('--ydim', default=10,type=int)
parser.add_argument('--gpu', default="6,7",type=str)
parser.add_argument('--anneal', default=False, action='store_true')
parser.add_argument('--PC', default=False,action='store_true')
parser.add_argument('--nobatchnorm', default=False,action='store_true')
parser.add_argument('--nofilewrite', default=False,action='store_true')
parser.add_argument('--noadjustlr', default=True,action='store_true')
parser.add_argument('--noadjustwd', default=True,action='store_true')
parser.add_argument('--noadjustbs', default=True,action='store_true')
parser.add_argument('--pMNIST', default=False,action='store_true')
parser.add_argument('--test', default=False,action='store_true')
parser.add_argument('--fileprefix', default='Equi',type=str)
parser.add_argument('--load', default=None,type=str)
parser.add_argument('--loadepoch', default=0,type=int)
parser.add_argument('--saveepoch', default=500,type=int)
parser.add_argument('--weightdecay', default=0., type=float)
parser.add_argument('--momentum', default=0., type=float)
parser.add_argument('--optswitch', default=0., type=float)

def Prepare():

    #torch.manual_seed(0)

    global args, Criterion, DS_train, DS_valid, DS_test, InitParaFile, LoadParaFile, LoadNetFile, Inst, MinInst

    args = parser.parse_args()

    args.lr = 2**(-args.loglr/4)

    if args.anneal == True:
        args.fileprefix = "anneal"
    else:
        args.fileprefix = "const"
    if args.PC==False:
        os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu
    #else:
    #    os.environ["CUDA_VISIBLE_DEVICES"] = "0"

    print(torch.cuda.current_device())

    Inst = args.instance
    MinInst = Inst*args.mininstance-Inst+1
    
    LoadParaFile=[]
    for i in range(MinInst,MinInst+Inst):
        ParaFileName="./parameters/InitPara_%d.pth" % args.initial
        l = -1
        while os.path.exists("./%s_parameters/Para_%s_log_%d_init_%d_epoch_%d_inst_%d.pt" % (args.fileprefix,args.fileprefix,args.loglr,args.initial,l+500,i+args.saveinst)):
            ParaFileName="./%s_parameters/Para_%s_log_%d_init_%d_epoch_%d_inst_%d.pt" % (args.fileprefix,args.fileprefix,args.loglr,args.initial,l+500,i+args.saveinst)
            l += 500	
        LoadParaFile.append((ParaFileName,l))
        print("Inst:",i,"Load:",l)

    LoadNetFile = "./nets/Net_%d.pth" % 0


Prepare()

def learn(i, LR = args.lr, EpochNumber = args.epochnumber, LoadEpoch = args.loadepoch):

    Vec = torch.load(LoadParaFile[i-MinInst][0])
    Vec.requires_grad = True
    Vec.retain_grad()
    
    Net=model.Net().cuda()
    Net.load_state_dict(torch.load(LoadNetFile))

    Output = torch.empty(0,4)
    weight_Output = torch.empty(0,Vec.size()[0])

    for epochcounter in range(LoadParaFile[i-MinInst][1]+1,EpochNumber):

        print("epoch ", epochcounter)

        TrainLoss, TrainGrad= 0, 0
        
        for batchcounter in range(args.epochlen):

            L, G = model.run(Vec,Net,LR,True)

            LR = args.lr + (LR - args.lr)/args.annealratio
                
            TrainLoss+=L.item()
            TrainGrad+=G.item()

        output = torch.empty(4)
        weight_output = torch.empty_like(Vec)
        
        TrainLoss /= args.epochlen
        TrainGrad /= args.epochlen

        output = torch.tensor([epochcounter+1, LR, \
                               TrainLoss, TrainGrad])
        weight_output = Vec
        
        Output = torch.cat((Output,output.unsqueeze(0))).detach()
        weight_Output = torch.cat((weight_Output,weight_output.unsqueeze(0).cpu())).detach()
        
        print(output)
        print(Output.size())
        print(weight_Output.size())

        if (epochcounter+1) % args.saveepoch == 0:
            OutputFile="./%s_results/Loss_%s_log_%d_init_%d_epoch_%d_inst_%d.pt" % (args.fileprefix,args.fileprefix,args.loglr,args.initial,epochcounter,i)
            weight_OutputFile="./%s_results/Weight_%s_log_%d_init_%d_epoch_%d_inst_%d.pt" % (args.fileprefix,args.fileprefix,args.loglr,args.initial,epochcounter,i)
            torch.save(Output,OutputFile)
            torch.save(weight_Output,weight_OutputFile)

            Output = torch.empty(0,4)
            weight_Output = torch.empty(0,Vec.size()[0])
            
            ParaFile="./%s_parameters/Para_%s_log_%d_init_%d_epoch_%d_inst_%d.pt" % (args.fileprefix,args.fileprefix,args.loglr,args.initial,epochcounter,i)
            torch.save(Vec, ParaFile)
    
    
for i in range(MinInst, MinInst+Inst):
    if args.anneal == True:
        learn(i,64)
    else:
        learn(i)
